//LIBBERYLSETTINGS
//SO VERSION 0
//see spec in header
#define _GNU_SOURCE
#include <stdlib.h>
#include <beryl-settings.h>
#include <string.h>
#include <malloc.h>
#include <libintl.h>
#include <dlfcn.h>
#include <beryl.h>

CompDisplay *display = NULL;

#define NEW(a,b) \
    a * b = malloc(sizeof(a)); memset((b),0,sizeof(a))

struct _Modifier {
    gchar *name;
    gint  modifier;
} const static modifiers[] = {
    { "<Shift>",      ShiftMask		 },
    { "<Control>",    ControlMask	 },
    { "<Mod1>",	      Mod1Mask		 },
    { "<Mod2>",	      Mod2Mask		 },
    { "<Mod3>",	      Mod3Mask		 },
    { "<Mod4>",	      Mod4Mask		 },
    { "<Mod5>",	      Mod5Mask		 },
    { "<Alt>",	      CompAltMask        },
    { "<Meta>",	      CompMetaMask       },
    { "<Super>",      CompSuperMask      },
    { "<Hyper>",      CompHyperMask	 },
    { "<ModeSwitch>", CompModeSwitchMask },
};

#define N_MODIFIERS (sizeof (modifiers) / sizeof (struct _Modifier))

unsigned int beryl_settings_get_mods_and_endptr(gchar * src, gchar ** ret)
{
    unsigned int mods=0;
    gchar * spos=src;
    while((spos=strchr(spos,'<')) && *src)
    {
        int i;
        for (i=0;i<N_MODIFIERS;i++)
            if (strncasecmp(modifiers[i].name,
                        spos,strlen(modifiers[i].name))==0)
            {
                mods|=modifiers[i].modifier;
                spos+=strlen(modifiers[i].name);
                src=spos;
                break;
            }
        if (i==N_MODIFIERS)
            break;
    }
    *ret=src;
    return mods;
}

gchar * beryl_settings_mods_to_string(unsigned int mods)
{
    gchar * retstr=g_strdup("");
    gchar * tmpstr=retstr;
    gint i;
    for (i=0;i<N_MODIFIERS;i++)
    {
        if (mods & modifiers[i].modifier)
        {
            retstr=g_strconcat(retstr,modifiers[i].name,NULL);
            g_free(tmpstr);
            tmpstr=retstr;
        }
    }
    return retstr;
}

//STATIC FUNCTIONS
static void copy_value(BerylSettingValue * dest, BerylSettingValue * src)
{
    BerylSettingType type;
    if (src->is_list_child)
        type=src->parent->info.for_list.list_of_type;
    else
        type=src->parent->type;
    dest->parent=src->parent;
    dest->is_list_child=src->is_list_child;
    if (type != BERYL_SETTING_TYPE_LIST)
    {
        if (type != BERYL_SETTING_TYPE_STRING)
            memcpy(&dest->value,&src->value,sizeof(BerylSettingValueUnion));
        else
            dest->value.as_string=g_strdup(src->value.as_string);
    }
    else
    {
        GSList * newsrc;
        dest->value.as_list=NULL;
        for (newsrc=src->value.as_list;newsrc;newsrc=newsrc->next)
        {
            NEW(BerylSettingValue,newdest);
            copy_value(newdest,newsrc->data);
            dest->value.as_list=g_slist_append(dest->value.as_list,newdest);
        }
    }
}

static void copy_from_default(BerylSetting * setting)
{
    copy_value(&setting->value,&setting->default_value);
}


static void init_info(BerylSettingInfo * info, CompOption * o, BerylSettingType type,
        gchar * gettext_domain)
{
    switch(type)
    {
        case BERYL_SETTING_TYPE_INT:
            info->for_int.min=o->rest.i.min;
            info->for_int.max=o->rest.i.max;
            break;
        case BERYL_SETTING_TYPE_FLOAT:
            info->for_float.min=o->rest.f.min;
            info->for_float.max=o->rest.f.max;
            info->for_float.precision=o->rest.f.precision;
            break;
        case BERYL_SETTING_TYPE_STRING:
            if (o->rest.s.nString && o->rest.s.string)
            {
                int i;
                for (i=0;i<o->rest.s.nString;i++)
                {
                    info->for_string.raw_values=g_slist_append(
                            info->for_string.raw_values,g_strdup(o->rest.s.string[i]));
                    info->for_string.allowed_values=g_slist_append(
                            info->for_string.allowed_values,
                            g_strdup(dgettext(gettext_domain,o->rest.s.string[i])));
                }
            }
            break;
        case BERYL_SETTING_TYPE_BINDING:
            info->for_bind.key=(o->value.action.state & 
                    CompActionStateInitKey) && TRUE;
            info->for_bind.button=(o->value.action.state & 
                    CompActionStateInitButton) && TRUE;
            info->for_bind.edge=(o->value.action.state & 
                    CompActionStateInitEdge) && TRUE;
            info->for_bind.bell=(o->value.action.state & 
                    CompActionStateInitBell) && TRUE;
            break;
        default:
            break;
    }
}

static void init_value(BerylSettingValue * value, CompOptionValue * from, BerylSettingType type,
        gchar * gettext_domain)
{
    switch(type)
    {
        case BERYL_SETTING_TYPE_INT:
            value->value.as_int=from->i;
            break;
        case BERYL_SETTING_TYPE_FLOAT:
            value->value.as_float=from->f;
            break;
        case BERYL_SETTING_TYPE_BOOL:
            value->value.as_bool=from->b;
            break;
        case BERYL_SETTING_TYPE_COLOR:
            {
                int i;
                for (i=0;i<4;i++)
                    value->value.as_color.array.array[i]=from->c[i];
            }
            break;
        case BERYL_SETTING_TYPE_STRING:
            value->value.as_string=g_strdup(from->s);
            break;
        case BERYL_SETTING_TYPE_BINDING:
            value->value.as_binding.enabled.value.key=(from->action.type & 
                    CompBindingTypeKey)?TRUE:FALSE;
            value->value.as_binding.enabled.value.button=(from->action.type & 
                    CompBindingTypeButton)?TRUE:FALSE;
            value->value.as_binding.button=from->action.button.button;
            value->value.as_binding.button_mod_mask=from->action.button.modifiers;
            value->value.as_binding.keysym=from->action.key.keysym;
            value->value.as_binding.key_mod_mask=from->action.key.modifiers;
            value->value.as_binding.on_bell=from->action.bell;
            value->value.as_binding.edge_mask=from->action.edgeMask;
            break;
        default:
            break;
    }
}

static BerylSettingType translate_type(CompOptionType type)
{
    switch(type)
    {
        case CompOptionTypeInt:
            return BERYL_SETTING_TYPE_INT;
        case CompOptionTypeFloat:
            return BERYL_SETTING_TYPE_FLOAT;
        case CompOptionTypeString:
            return BERYL_SETTING_TYPE_STRING;
        case CompOptionTypeColor:
            return BERYL_SETTING_TYPE_COLOR;
        case CompOptionTypeBool:
            return BERYL_SETTING_TYPE_BOOL;
        case CompOptionTypeAction:
            return BERYL_SETTING_TYPE_BINDING;
        case CompOptionTypeList:
            return BERYL_SETTING_TYPE_LIST;
        default:
            break;
    }
    return BERYL_SETTING_TYPE_COUNT;
}

static void init_option(BerylSettingsPlugin * plugin, CompOption * o, gboolean is_screen)
{
    NEW(BerylSetting,setting);
    setting->parent=plugin;
    setting->is_screen=is_screen;
    setting->is_default=TRUE;
    setting->name=g_strdup(o->name);
    setting->short_desc=g_strdup(
            dgettext(plugin->gettext_domain,o->shortDesc));
    setting->long_desc=g_strdup(dgettext(plugin->gettext_domain,o->longDesc));
    setting->group=g_strdup(strlen(o->group)?dgettext(plugin->gettext_domain,o->group):"");
    setting->subGroup=g_strdup(strlen(o->subGroup)?dgettext(plugin->gettext_domain,o->subGroup):"");
    setting->displayHints=g_strdup(o->displayHints);
    setting->type=translate_type(o->type);
    setting->default_value.parent=setting;
    if (setting->type==BERYL_SETTING_TYPE_LIST)
    {
        NEW(BerylSettingInfo,list_info);
        setting->info.for_list.list_of_type=translate_type(o->value.list.type);
        setting->info.for_list.list_of_info=list_info;
        init_info(setting->info.for_list.list_of_info,o,
                setting->info.for_list.list_of_type,plugin->gettext_domain);
        int i;
        for (i=0;i<o->value.list.nValue;i++)
        {
            NEW(BerylSettingValue,val);
            val->parent=setting;
            val->is_list_child=TRUE;
            init_value(val,&o->value.list.value[i],setting->info.for_list.list_of_type,
                    plugin->gettext_domain);
            setting->default_value.value.as_list=
                g_slist_append(setting->default_value.value.as_list,val);
        }
    }
    else
    {
        init_info(&setting->info,o,setting->type,plugin->gettext_domain);
        init_value(&setting->default_value,&o->value,setting->type,plugin->gettext_domain);
    }
    plugin->settings=g_slist_append(plugin->settings,setting);
    copy_from_default(setting);
}

static void comp_value_to_setting_value(CompOptionValue * comp_value, BerylSettingValue * beryl_value)
{
    BerylSettingType type;
    if (beryl_value->is_list_child)
        type=beryl_value->parent->info.for_list.list_of_type;
    else
        type=beryl_value->parent->type;
    beryl_value->parent->is_default=FALSE;
    switch(type)
    {
        case BERYL_SETTING_TYPE_INT:
            beryl_value->value.as_int=comp_value->i;
            break;
        case BERYL_SETTING_TYPE_FLOAT:
            beryl_value->value.as_float=comp_value->f;
            break;
        case BERYL_SETTING_TYPE_BOOL:
            beryl_value->value.as_bool=comp_value->b;
            break;
        case BERYL_SETTING_TYPE_COLOR:
            {
                int i;
                for (i=0;i<4;i++)
                    beryl_value->value.as_color.array.array[i]=comp_value->c[i];
            }
            break;
        case BERYL_SETTING_TYPE_STRING:
            if (beryl_value->value.as_string)
                free(beryl_value->value.as_string);
            beryl_value->value.as_string=g_strdup(comp_value->s);
            break;
        case BERYL_SETTING_TYPE_BINDING:
            beryl_value->value.as_binding.enabled.value.key =
                (comp_value->action.type & CompBindingTypeKey) != 0;
            beryl_value->value.as_binding.key_mod_mask = comp_value->action.key.modifiers;
            beryl_value->value.as_binding.keysym = comp_value->action.key.keysym;
            beryl_value->value.as_binding.enabled.value.button =
                (comp_value->action.type & CompBindingTypeButton) != 0;
            beryl_value->value.as_binding.button_mod_mask = comp_value->action.button.modifiers;
            beryl_value->value.as_binding.button = comp_value->action.button.button;
            beryl_value->value.as_binding.edge_mask = comp_value->action.edgeMask;
            beryl_value->value.as_binding.on_bell = comp_value->action.bell;
            break;
        case BERYL_SETTING_TYPE_LIST:
            {
                int i;
                beryl_setting_list_clear(beryl_value->parent);
                for (i=0;i<comp_value->list.nValue;i++)
                {
                    BerylSettingValue * val = beryl_setting_list_append(beryl_value->parent);
                    comp_value_to_setting_value(&comp_value->list.value[i],val);
                }
            }
            break;
        default:
            break;
    }

}

static void setting_value_to_comp_value(BerylSettingValue * beryl_value, CompOptionValue * comp_value)
{
    BerylSettingType type;
    if (beryl_value->is_list_child)
        type=beryl_value->parent->info.for_list.list_of_type;
    else
        type=beryl_value->parent->type;
    switch(type)
    {
        case BERYL_SETTING_TYPE_INT:
            comp_value->i=beryl_value->value.as_int;
            break;
        case BERYL_SETTING_TYPE_FLOAT:
            comp_value->f=beryl_value->value.as_float;
            break;
        case BERYL_SETTING_TYPE_BOOL:
            comp_value->b=beryl_value->value.as_bool;
            break;
        case BERYL_SETTING_TYPE_COLOR:
            {
                int i;
                for (i=0;i<4;i++)
                    comp_value->c[i]=beryl_value->value.as_color.array.array[i];
            }
            break;
        case BERYL_SETTING_TYPE_STRING:
            comp_value->s=strdup(beryl_value->value.as_string);
            break;
        case BERYL_SETTING_TYPE_BINDING:
            comp_value->action.type = CompBindingTypeNone;
            if (beryl_value->value.as_binding.enabled.value.key)
                comp_value->action.type |= CompBindingTypeKey;
            comp_value->action.key.modifiers = beryl_value->value.as_binding.key_mod_mask;
            comp_value->action.key.keysym = beryl_value->value.as_binding.keysym;
            if (beryl_value->value.as_binding.enabled.value.button)
                comp_value->action.type |= CompBindingTypeButton;
            comp_value->action.button.modifiers = beryl_value->value.as_binding.button_mod_mask;
            comp_value->action.button.button = beryl_value->value.as_binding.button;
            comp_value->action.edgeMask=beryl_value->value.as_binding.edge_mask;
            comp_value->action.bell=beryl_value->value.as_binding.on_bell;
            break;
        case BERYL_SETTING_TYPE_LIST:
            {
                int len=g_slist_length(beryl_value->value.as_list);
                comp_value->list.nValue=len;
                if (len)
                {
                    comp_value->list.value=malloc(sizeof(CompOptionValue)*len);
                    memset(comp_value->list.value,0,
                            sizeof(CompOptionValue)*len);
                    GSList * i;
                    int j=0;
                    for (i=beryl_value->value.as_list;i;i=i->next)
                    {
                        setting_value_to_comp_value(i->data,&comp_value->list.value[j]);
                        j++;
                    }
                }
            }
            break;
        default:
            break;
    }
}

typedef struct _FindSubGroup
{
    gchar * subGroup;
    gboolean found;
    BerylSettingsSubGroup * found_at;
} FindSubGroup;

static void find_subgroup(BerylSettingsSubGroup * subGroup, FindSubGroup * find)
{
    if (find->found) return;
    if (!strcmp(subGroup->name,find->subGroup))
    {
	find->found=TRUE;
	find->found_at=subGroup;
    }
}

static void subgroup_add(BerylSetting * setting, BerylSettingsGroup * group)
{
    FindSubGroup f;
    f.found=FALSE;
    f.subGroup=setting->subGroup;
    g_slist_foreach(group->subGroups,(GFunc)find_subgroup,&f);
    if (!f.found)
    {
	NEW(BerylSettingsSubGroup,g);
	group->subGroups=g_slist_append(group->subGroups,g);
	g->name=g_strdup(setting->subGroup);
	f.found_at=g;
    }
    f.found_at->settings=g_slist_append(f.found_at->settings,setting);
}

typedef struct _FindGroup
{
    gchar * group;
    gboolean found;
    BerylSettingsGroup * found_at;
} FindGroup;

static void find_group(BerylSettingsGroup * group, FindGroup * find)
{
    if (find->found) return;
    if (!strcmp(group->name,find->group))
    {
	find->found=TRUE;
	find->found_at=group;
    }
}

static void group_add(BerylSetting * setting, BerylSettingsPlugin * plugin)
{
    FindGroup f;
    f.found=FALSE;
    f.group=setting->group;
    g_slist_foreach(plugin->groups,(GFunc)find_group,&f);
    if (!f.found)
    {
	NEW(BerylSettingsGroup,g);
	plugin->groups=g_slist_append(plugin->groups,g);
	g->name=g_strdup(setting->group);
	f.found_at=g;
    }
    subgroup_add(setting,f.found_at);
}

static void collate_groups(BerylSettingsPlugin * plugin)
{
    g_slist_foreach(plugin->settings,(GFunc)group_add,plugin);
}

static void load_plugin(BerylSettingsContext * context, gchar * filename)
{
    void * dlhand;
    gchar * err;
    PluginGetInfoProc getInfo;
    CompPluginVTable * vt;
    dlhand=dlopen(filename,RTLD_LAZY);
    if (!dlhand)
    {
        err=dlerror();
        if (err)
            fprintf(stderr, _("libberylsettings: dlopen: %s\n"),err);
        return;
    }
    dlerror();
    getInfo = (PluginGetInfoProc) dlsym (dlhand, "getCompPluginInfo");
    if ((err=dlerror()))
    {
        fprintf (stderr, _("libberylsettings: dlsym: %s\n"), err);
        dlclose(dlhand);
        return;
    }
    vt=getInfo();
    if (vt->version != BERYL_VERSION ||
            vt->struct_plugin_size != sizeof (CompPlugin) ||
            vt->struct_display_size != sizeof (CompDisplay) ||
            vt->struct_screen_size != sizeof (CompScreen) ||
            vt->struct_window_size != sizeof (CompWindow) ||
            vt->struct_texture_size != sizeof (CompTexture) ||
            vt->struct_icon_size != sizeof (CompIcon))
    {
        fprintf (stderr, _("libberylsettings: Couldn't get vtable from '%s' plugin\n"), filename);
        dlclose (dlhand);
        return;
    }
    //make sure a plugin by the same name isn't already loaded
    if (beryl_settings_context_find_plugin(context,vt->name))
    {
        dlclose(dlhand);
        return;
    }
    NEW(BerylSettingsPlugin,plugin);
    plugin->name=g_strdup(vt->name);
    plugin->gettext_domain=g_strdup(vt->gettext_domain);
    plugin->short_desc=g_strdup(dgettext(plugin->gettext_domain,vt->shortDesc));
    plugin->long_desc=g_strdup(dgettext(plugin->gettext_domain,vt->longDesc));
    int n;
    for (n=0;n<vt->nDeps;n++)
    {
        switch(vt->deps[n].rule)
        {
            case CompPluginRuleBefore:
                plugin->load_before=g_slist_append(
                        plugin->load_before,g_strdup(vt->deps[n].name));
                break;
            case CompPluginRuleAfter:
                plugin->load_after=g_slist_append(
                        plugin->load_after,g_strdup(vt->deps[n].name));
                break;
            case CompPluginRuleRequire:
                plugin->requires=g_slist_append(
                        plugin->requires,g_strdup(vt->deps[n].name));
                break;
            default:
                break;
        }
    }
    for (n=0;n<vt->nFeatures;n++)
    {
        plugin->provides=g_slist_append(plugin->provides,
                g_strdup(vt->features[n].name));
    }
    if (vt->getDisplayOptions)
    {
        int n;
        CompOption * o;
        o = vt->getDisplayOptions(NULL,&n);
        for (;n;n--)
        {
            init_option(plugin,o,FALSE);
            o++;
        }
    }
    if (vt->getScreenOptions)
    {
        int n;
        CompOption * o;
        o = vt->getScreenOptions(NULL,&n);
        for (;n;n--)
        {
            init_option(plugin,o,TRUE);
            o++;
        }
    }
    dlclose(dlhand);
    collate_groups(plugin);
    context->plugins=g_slist_append(context->plugins,plugin);
}

static void load_plugins(BerylSettingsContext * context, gchar * path)
/* iterates through $path and adds any plugins found to the context
            $context        the context to add to
            $path           the path to search for plugins (any .so file) */
{
    GDir * d;
    const gchar * n;
    d=g_dir_open(path,0,NULL);
    if (!d)
        return; // no such path
    while((n=g_dir_read_name(d)))
    {
        if (g_str_has_suffix(n,".so") && g_str_has_prefix(n,"lib"))
        {
            gchar * p;
            p=g_strconcat(path,"/",n,NULL);
            //actually load the plugin
            load_plugin(context,p);
            g_free(p);
        }
    }
    g_dir_close(d);
}

static void init_core_settings (BerylSettingsContext * context)
{
    NEW(BerylSettingsPlugin,plugin);
    int n;
    CompOption * o;
    display = malloc(sizeof(CompDisplay));
    compDisplayInitOptions(display,NULL,0);
    plugin->gettext_domain=g_strdup("beryl-core");
    o=compGetDisplayOptions(NULL,&n);
    for (;n;n--)
    {
        init_option(plugin,o,FALSE);
        o++;
    }
    o=compGetOptions(&n);
    for (;n;n--)
    {
        init_option(plugin,o,TRUE);
        o++;
    }
    collate_groups(plugin);
    context->plugins=g_slist_append(context->plugins,plugin);
}

static void free_value(BerylSettingValue * value)
{
    BerylSettingType type;
    if (value->is_list_child)
        type=value->parent->info.for_list.list_of_type;
    else
        type=value->parent->type;
    switch (type)
    {
        case BERYL_SETTING_TYPE_LIST:
            g_slist_foreach(value->value.as_list,
                    (GFunc)free_value,NULL);
            if (value->value.as_list)
                g_slist_free(value->value.as_list);
            break;
        case BERYL_SETTING_TYPE_STRING:
            if (value->value.as_string)
                g_free(value->value.as_string);
        default:
            break;
    }
    if (value->is_list_child)
        free(value);
}

static void free_string_info(BerylSettingInfo * info)
{
    g_slist_foreach(info->for_string.allowed_values,
            (GFunc)g_free,NULL);
    g_slist_foreach(info->for_string.raw_values,
            (GFunc)g_free,NULL);
    if (info->for_string.allowed_values)
        g_slist_free(info->for_string.allowed_values);
    if (info->for_string.raw_values)
        g_slist_free(info->for_string.raw_values);
}

static void free_setting(BerylSetting * setting)
{
    if (setting->name) g_free(setting->name);
    if (setting->short_desc) g_free(setting->short_desc);
    if (setting->long_desc) g_free(setting->long_desc);
    if (setting->group) g_free(setting->group);
    if (setting->subGroup) g_free(setting->subGroup);
    if (setting->displayHints) g_free(setting->displayHints);
    free_value(&setting->default_value);
    free_value(&setting->value);
    if (setting->type == BERYL_SETTING_TYPE_STRING)       
        free_string_info(&setting->info);
    if (setting->type == BERYL_SETTING_TYPE_LIST &&
            setting->info.for_list.list_of_type == BERYL_SETTING_TYPE_STRING)
        free_string_info(setting->info.for_list.list_of_info);
    if (setting->type == BERYL_SETTING_TYPE_LIST)
        free(setting->info.for_list.list_of_info);
    free(setting);
}

static void free_subgroup(BerylSettingsSubGroup * subGroup)
{
    g_free(subGroup->name);
    if (subGroup->settings) g_slist_free(subGroup->settings);
}

static void free_group(BerylSettingsGroup * group)
{
    g_slist_foreach(group->subGroups,(GFunc)free_subgroup,NULL);
    if (group->subGroups) g_slist_free(group->subGroups);
    g_free(group->name);
}

static void free_plugin(BerylSettingsPlugin * plugin)
{
    if (plugin->name) g_free(plugin->name);
    if (plugin->short_desc) g_free(plugin->short_desc);
    if (plugin->long_desc) g_free(plugin->long_desc);
    if (plugin->gettext_domain) g_free(plugin->gettext_domain);
    g_slist_foreach(plugin->load_after,(GFunc)g_free,NULL);
    g_slist_foreach(plugin->load_before,(GFunc)g_free,NULL);
    g_slist_foreach(plugin->provides,(GFunc)g_free,NULL);
    g_slist_foreach(plugin->requires,(GFunc)g_free,NULL);
    g_slist_foreach(plugin->settings,(GFunc)free_setting,NULL);
    g_slist_foreach(plugin->groups,(GFunc)free_group,NULL);
    if (plugin->groups) g_slist_free(plugin->groups);
    if (plugin->load_after) g_slist_free(plugin->load_after);
    if (plugin->load_before) g_slist_free(plugin->load_before);
    if (plugin->provides) g_slist_free(plugin->provides);
    if (plugin->requires) g_slist_free(plugin->requires);
    if (plugin->settings) g_slist_free(plugin->settings);
    free(plugin);
}

typedef struct _BerylSettingsFindPlugin
{
    gchar * search_for;
    BerylSettingsPlugin * found;
} BerylSettingsFindPlugin;

static void find_plugin(BerylSettingsPlugin * plugin, BerylSettingsFindPlugin * find)
{
    if (find->found)
        return;
    if (find->search_for==NULL && plugin->name==NULL)
    {
        find->found=plugin;
        return;
    }
    if (find->search_for && plugin->name && strcmp(find->search_for,plugin->name)==0)
        find->found=plugin;
}

typedef struct _BerylSettingsFindSetting
{
    gchar * search_for;
    gboolean is_screen;
    BerylSetting * found;
} BerylSettingsFindSetting;

static void find_setting(BerylSetting * setting, BerylSettingsFindSetting * find)
{
    if (find->found)
        return;
    if (find->search_for && 
            setting->name && 
            find->is_screen==setting->is_screen &&
            strcmp(find->search_for,setting->name)==0)
        find->found=setting;
}
static void read_setting(BerylSetting * setting, BerylSettingsContext * c)
{
    if (c->read_setting)
	c->read_setting(c,setting);
}
static void write_setting(BerylSetting * setting, BerylSettingsContext * c)
{
    if (c->write_setting)
	c->write_setting(c,setting);
}
static void read_settings_for_plugin(BerylSettingsPlugin * plugin, 
	BerylSettingsContext * c)
{
    g_slist_foreach(plugin->settings,(GFunc)read_setting,c);
}

static void write_settings_for_plugin(BerylSettingsPlugin * plugin, 
	BerylSettingsContext * c)
{
    g_slist_foreach(plugin->settings,(GFunc)write_setting,c);
}

typedef struct _BerylFindListSettingIndex
{
    int index;
    gboolean found;
    BerylSettingValue * value;
} BerylFindListSettingIndex;

static void find_index_of(BerylSettingValue * value, BerylFindListSettingIndex * find)
{
    if (find->found)
        return;
    if (find->value==value)
        find->found=TRUE;
    else
        find->index++;
}

static gboolean check_s_type(BerylSetting * setting, BerylSettingType type)
{
    if (setting->type==BERYL_SETTING_TYPE_LIST)
        return setting->info.for_list.list_of_type==type;
    else
        return setting->type==type;
}

static BerylSettingInfo * get_s_info(BerylSetting * setting)
{
    if (setting->type==BERYL_SETTING_TYPE_LIST)
        return setting->info.for_list.list_of_info;
    else
        return &setting->info;
}

static gboolean check_type(BerylSettingValue * value, BerylSettingType type)
{
    if (value->is_list_child)
        return value->parent->info.for_list.list_of_type==type;
    else
        return value->parent->type==type;
}

static BerylSettingInfo * get_info(BerylSettingValue * value)
{
    if (value->is_list_child)
        return value->parent->info.for_list.list_of_info;
    else
        return &value->parent->info;
}

//EXPORTED FUNCTIONS
BerylSettingsContext * beryl_settings_context_new(void)
{
    gchar * homeplugins=g_strconcat(g_get_home_dir(),"/.beryl/plugins",NULL);
    NEW(BerylSettingsContext,context);
    init_core_settings(context);
    load_plugins(context,homeplugins);
    load_plugins(context,PLUGINDIR);    // load after, because that way name conflicts
                                        // will lead to the home version being loaded
    g_free(homeplugins);
    beryl_settings_context_set_backend(context,"ini");
    return context;
}

gboolean beryl_settings_context_set_backend(BerylSettingsContext * context,
	gchar * backend)
{
    if (context->backend)
    {
	dlclose(context->backend_dlhand);
	context->backend_dlhand=0;
	context->read_setting=0;
	context->write_setting=0;
	context->read_init=0;
	context->read_done=0;
	context->write_init=0;
	context->write_done=0;
	g_free(context->backend);
	context->backend=0;
    }
    // prefer local backend...
    gchar * dlname=g_strconcat(g_get_home_dir(),"/.beryl/backends/lib",
	    backend,".so",NULL);
    gchar * eret;
    gchar * eret2=NULL;
    dlerror();
    context->backend_dlhand=dlopen(dlname,RTLD_NOW);
    eret=dlerror();
    if (eret || !context->backend_dlhand)
    {
	eret2=eret;
	g_free(dlname);
	dlname=g_strconcat(PLUGINDIR,"/backends/lib",backend,".so",NULL);
	context->backend_dlhand=dlopen(dlname,RTLD_NOW);
	eret=dlerror();
    }
    if (eret || !context->backend_dlhand)
    {
	if (eret2)
	    g_message(eret2);
	g_message(eret);
	g_free(dlname);
	return FALSE;
    }
    context->backend=g_strdup(backend);
#define LOAD_ITEM(n) \
    context->n =dlsym(context->backend_dlhand,#n);\
    if(!context->n  )\
    	g_warning(#n " not found in backend %s",backend);
    LOAD_ITEM(read_setting);
    LOAD_ITEM(write_setting);
    LOAD_ITEM(read_init);
    LOAD_ITEM(read_done);
    LOAD_ITEM(write_init);
    LOAD_ITEM(write_done);
    return TRUE;
}

void beryl_settings_context_destroy(BerylSettingsContext * context)
{
    g_slist_foreach(context->plugins,(GFunc)free_plugin,NULL);
    if (context->plugins) g_slist_free(context->plugins);
    if (context->backend)
    {
	dlclose(context->backend_dlhand);
	g_free(context->backend);
    }
    free(context);
}

GSList * beryl_settings_plugin_get_groups(BerylSettingsPlugin * plugin)
{
    return plugin->groups;
}

gchar * beryl_settings_group_get_name(BerylSettingsGroup * group)
{
    return group->name;
}

GSList * beryl_settings_group_get_subgroups(BerylSettingsGroup * group)
{
    return group->subGroups;
}

gchar * beryl_settings_subgroup_get_name(BerylSettingsSubGroup * subGroup)
{
    return subGroup->name;
}

GSList * beryl_settings_subgroup_get_settings(BerylSettingsSubGroup * subGroup)
{
    return subGroup->settings;
}

gboolean beryl_setting_has_hint(BerylSetting * setting, gchar * hint)
{
    gboolean ret;
    gchar * hhint = g_strconcat(hint,";",NULL);
    ret=strstr(setting->displayHints,hhint)?TRUE:FALSE;
    g_free(hhint);
    return ret;
}

BerylSettingsPlugin * beryl_settings_context_find_plugin(BerylSettingsContext * context, gchar * name)
{
    BerylSettingsFindPlugin find;
    find.found=NULL;
    find.search_for=name;
    g_slist_foreach(context->plugins,(GFunc)find_plugin,&find);
    return find.found;
}

BerylSetting * beryl_settings_plugin_find_setting(BerylSettingsPlugin * plugin, gchar * name,
        gboolean is_screen)
{
    BerylSettingsFindSetting find;
    find.found=NULL;
    find.search_for=name;
    find.is_screen=is_screen;
    g_slist_foreach(plugin->settings,(GFunc)find_setting,&find);
    return find.found;
}

void beryl_settings_context_read(BerylSettingsContext * context)
{
    if (context->read_init)
	if (!context->read_init(context)) return;
    g_slist_foreach(context->plugins,(GFunc)read_settings_for_plugin,context);
    if (context->read_done)
	context->read_done(context);
}

void beryl_settings_context_write(BerylSettingsContext * context)
{
    if (context->write_init)
	if (!context->write_init(context)) return;
    g_slist_foreach(context->plugins,(GFunc)write_settings_for_plugin,context);
    if (context->write_done)
	context->write_done(context);
}

BerylSettingsContext * beryl_settings_context_comp_new(void)
{
    NEW(BerylSettingsContext,context);
    context->is_comp=TRUE;
    beryl_settings_context_set_backend(context,"ini");
    return context;
}

gboolean beryl_settings_context_comp_add_plugin(BerylSettingsContext * context, gchar * name)
{
    if (!context->is_comp) return FALSE;
    if (beryl_settings_context_find_plugin(context,name))
        return FALSE;
    NEW(BerylSettingsPlugin,plugin);
    if (name)
        plugin->name=g_strdup(name);
    plugin->gettext_domain=g_strdup("");
    context->plugins=g_slist_append(context->plugins,plugin);
    return TRUE;
}

#if 0 /* Incomplete */
void beryl_settings_context_comp_free_option(s_plugin, option)
{
	BerylSetting * setting;

	g_free(setting->name);
	g_free(setting->short_desc);
	g_free(setting->long_desc);
	free_value(setting->value);
}
#endif

gboolean beryl_settings_context_comp_add_option(BerylSettingsContext * context,
        gchar * plugin, CompOption * option, gboolean is_screen)
{
    if (!context->is_comp) return FALSE;
    BerylSettingsPlugin * s_plugin;
    if (!(s_plugin=beryl_settings_context_find_plugin(context,plugin)))
        return FALSE;
    if (beryl_settings_plugin_find_setting(s_plugin,option->name,is_screen))
        return FALSE;
    init_option(s_plugin,option,is_screen);
    return TRUE;
}

gboolean beryl_settings_context_comp_get_option_value(BerylSettingsContext * context,
        gchar * plugin, gchar * name, gboolean is_screen, CompOptionValue * value)
{
    if (!context->is_comp) return FALSE;
    BerylSettingsPlugin * s_plugin;
    BerylSetting * setting;
    if (!(s_plugin=beryl_settings_context_find_plugin(context,plugin)))
    {
        return FALSE;
    }
    if (!(setting=beryl_settings_plugin_find_setting(s_plugin,name,is_screen)))
    {
        return FALSE;
    }
    setting_value_to_comp_value(&setting->value,value);
    return TRUE;
}

gboolean beryl_settings_context_comp_set_option_value(BerylSettingsContext * context,
        gchar * plugin, gchar * name, gboolean is_screen, CompOptionValue * value)
{
    if (!context->is_comp) return FALSE;
    BerylSettingsPlugin * s_plugin;
    BerylSetting * setting;
    if (!(s_plugin=beryl_settings_context_find_plugin(context,plugin)))
        return FALSE;
    if (!(setting=beryl_settings_plugin_find_setting(s_plugin,name,is_screen)))
        return FALSE;
    setting->is_default=FALSE;
    comp_value_to_setting_value(value,&setting->value);
    return TRUE;
}

gboolean beryl_settings_send_reload_signal(void)
{
    gchar * args[] =
    {"killall","-u",(gchar *)g_get_user_name(),"-SIGUSR1","-r","^beryl$|^beryl-xgl$",NULL};
    gchar * ret=NULL;
    if (!g_spawn_sync(NULL,args,NULL,G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_SEARCH_PATH,
                NULL,NULL,&ret,NULL,NULL,NULL) || !ret)
        return FALSE;
    return TRUE;
}

void beryl_setting_reset_to_default(BerylSetting * setting)
{
    //if (!setting->is_default)
    free_value(&setting->value);
    copy_from_default(setting);
    setting->is_default=TRUE;
}

void beryl_setting_list_clear(BerylSetting * setting)
{
    if (setting->type != BERYL_SETTING_TYPE_LIST)
        return;
    setting->is_default=FALSE;
    g_slist_foreach(setting->value.value.as_list,
            (GFunc)free_value,NULL);
    if (setting->value.value.as_list)
        g_slist_free(setting->value.value.as_list);
    setting->value.value.as_list=NULL;
}

BerylSettingValue * beryl_setting_list_append(BerylSetting * setting)
{
    if (setting->type != BERYL_SETTING_TYPE_LIST)
        return NULL;
    setting->is_default=FALSE;
    NEW(BerylSettingValue,value);
    value->parent=setting;
    value->is_list_child=TRUE;
    setting->value.value.as_list=g_slist_append(setting->value.value.as_list,value);
    return value;
}

gint beryl_setting_list_length(BerylSetting * setting)
{
    if (setting->type != BERYL_SETTING_TYPE_LIST)
        return -1;
    return g_slist_length(setting->value.value.as_list);
}

gint beryl_setting_list_value_index(BerylSettingValue * value)
{
    if (!value->is_list_child)
        return -1;
    BerylFindListSettingIndex find;
    find.found=FALSE;
    find.index=0;
    find.value=value;
    g_slist_foreach(value->parent->value.value.as_list,(GFunc)find_index_of,&find);
    if (find.found)
        return find.index;
    else
        return -1;
}

void beryl_setting_list_value_swap_with(BerylSettingValue * value, guint n)
{
    if (!value->is_list_child)
        return;
    GSList * a = g_slist_nth(value->parent->value.value.as_list,n);
    GSList * b = g_slist_find(value->parent->value.value.as_list,value);
    if (!a || !b)
        return;
    value->parent->is_default=FALSE;
    b->data=a->data;
    a->data=value;
}

void beryl_setting_list_value_move_before(BerylSettingValue * value, guint n)
{
    if (!value->is_list_child)
        return;
    GSList * a = g_slist_nth(value->parent->value.value.as_list,n);
    if (a && a->data==value)
        return;
    value->parent->is_default=FALSE;
    if (!a)
    {
        value->parent->value.value.as_list=g_slist_append(g_slist_remove(
                value->parent->value.value.as_list,value),value);
        return;
    }
    value->parent->value.value.as_list=g_slist_insert_before(
            g_slist_remove(value->parent->value.value.as_list,value),a,value);
}

void beryl_setting_list_value_remove(BerylSettingValue * value)
{
    if (!value->is_list_child)
        return;
    value->parent->is_default=FALSE;
    value->parent->value.value.as_list=g_slist_remove(
            value->parent->value.value.as_list,value);
    free_value(value);
}

BerylSettingValue * beryl_setting_get_primary_value(BerylSetting * setting)
{
    return &setting->value;
}

gboolean beryl_setting_value_get_int(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_INT))
        return FALSE;
    *data=value->value.as_int;
    return TRUE;
}

gboolean beryl_setting_value_get_float(BerylSettingValue * value, gdouble * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_FLOAT))
        return FALSE;
    *data=value->value.as_float;
    return TRUE;
}

gboolean beryl_setting_value_get_bool(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BOOL))
        return FALSE;
    *data=value->value.as_bool;
    return TRUE;
}

gboolean beryl_setting_value_get_string(BerylSettingValue * value, const gchar ** data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_STRING))
        return FALSE;
    BerylSettingInfo * info=get_info(value);
    if (info->for_string.raw_values)
    {

        GSList * reststring=info->for_string.raw_values;
        GSList * xlatstring=info->for_string.allowed_values;
        while(reststring)
        {
            if (strcmp(value->value.as_string,reststring->data)==0)
            {
                *data=xlatstring->data;
                return TRUE;
            }
            reststring=reststring->next;
            xlatstring=xlatstring->next;
        }
    }
    else
    {
        *data=value->value.as_string;
        return TRUE;
    }
    return FALSE;
}

gboolean beryl_setting_value_get_keysym(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.keysym;
    return TRUE;
}

gboolean beryl_setting_value_get_keymods(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.key_mod_mask;
    return TRUE;
}

gboolean beryl_setting_value_get_button(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.button;
    return TRUE;
}

gboolean beryl_setting_value_get_buttonmods(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.button_mod_mask;
    return TRUE;
}

gboolean beryl_setting_value_get_edgemask(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.edge_mask;
    return TRUE;
}

gboolean beryl_setting_value_get_bell(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.on_bell;
    return TRUE;
}

gboolean beryl_setting_value_get_key_enabled(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.enabled.value.key;
    return TRUE;
}

gboolean beryl_setting_value_get_button_enabled(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=value->value.as_binding.enabled.value.button;
    return TRUE;
}

gboolean beryl_setting_value_get_color(BerylSettingValue * value, BerylSettingColorValue * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_COLOR))
        return FALSE;
    memcpy(data,&value->value.as_color,sizeof(BerylSettingColorValue));
    return TRUE;
}

gboolean beryl_setting_value_get_value_list(BerylSettingValue * value, GSList ** data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_LIST))
        return FALSE;
    *data=value->value.as_list;
    return TRUE;
}

gboolean beryl_setting_value_set_int(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_INT))
        return FALSE;
    BerylSettingInfo * info = get_info(value);
    if (*data<info->for_int.min || *data>info->for_int.max)
        return FALSE;
    value->parent->is_default=FALSE;
    value->value.as_int=*data;
    return TRUE;
}

gboolean beryl_setting_value_set_float(BerylSettingValue * value, gdouble * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_FLOAT))
        return FALSE;
    BerylSettingInfo * info = get_info(value);
    if (*data<info->for_float.min || *data>info->for_float.max)
        return FALSE;
    value->parent->is_default=FALSE;
    value->value.as_float=*data;
    return TRUE;
}

gboolean beryl_setting_value_set_bool(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BOOL))
        return FALSE;
    value->parent->is_default=FALSE;
    value->value.as_bool=*data;
    return TRUE;
}

gboolean beryl_setting_value_set_string(BerylSettingValue * value, const gchar ** data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_STRING))
        return FALSE;
    BerylSettingInfo * info = get_info(value);
    if (info->for_string.raw_values)
    {
        GSList * reststring=info->for_string.raw_values;
        GSList * xlatstring=info->for_string.allowed_values;
        while(reststring)
        {
            if (strcmp(*data,xlatstring->data)==0)
            {
                if (value->value.as_string)
                    g_free(value->value.as_string);
                value->value.as_string=g_strdup(reststring->data);
                value->parent->is_default=FALSE;
                return TRUE;
            }
            reststring=reststring->next;
            xlatstring=xlatstring->next;
        }
    }
    else
    {
        if (value->value.as_string)
            g_free(value->value.as_string);
        value->value.as_string=g_strdup(*data);
        value->parent->is_default=FALSE;
        return TRUE;
    }
    return FALSE;
}

gboolean beryl_setting_value_set_keysym(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.key)
        return FALSE;
    value->value.as_binding.keysym=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_keymods(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.key)
        return FALSE;
    value->value.as_binding.key_mod_mask=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_button(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.button)
        return FALSE;
    value->value.as_binding.button=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_buttonmods(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.button)
        return FALSE;
    value->value.as_binding.button_mod_mask=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_edgemask(BerylSettingValue * value, gint * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.edge)
        return FALSE;
    value->value.as_binding.edge_mask=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_bell(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.bell)
        return FALSE;
    value->value.as_binding.on_bell=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_key_enabled(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.key)
        return FALSE;
    value->value.as_binding.enabled.value.key=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_button_enabled(BerylSettingValue * value, gboolean * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    if (!get_info(value)->for_bind.button)
        return FALSE;
    value->value.as_binding.enabled.value.button=*data;
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_value_set_color(BerylSettingValue * value, BerylSettingColorValue * data)
{
    if (!check_type(value,BERYL_SETTING_TYPE_COLOR))
        return FALSE;
    memcpy(&value->value.as_color,data,sizeof(BerylSettingColorValue));
    value->parent->is_default=FALSE;
    return TRUE;
}

gboolean beryl_setting_get_can_set_key(BerylSetting * setting, gboolean * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=get_s_info(setting)->for_bind.key;
    return TRUE;
}

gboolean beryl_setting_get_can_set_button(BerylSetting * setting, gboolean * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=get_s_info(setting)->for_bind.button;
    return TRUE;
}

gboolean beryl_setting_get_can_set_edgemask(BerylSetting * setting, gboolean * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=get_s_info(setting)->for_bind.edge;
    return TRUE;
}

gboolean beryl_setting_get_can_set_bell(BerylSetting * setting, gboolean * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_BINDING))
        return FALSE;
    *data=get_s_info(setting)->for_bind.bell;
    return TRUE;
}

gboolean beryl_setting_get_allowed_strings(BerylSetting * setting, GSList ** data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_STRING))
        return FALSE;
    *data=get_s_info(setting)->for_string.allowed_values;
    return TRUE;
}

gboolean beryl_setting_get_raw_allowed_strings(BerylSetting * setting, GSList ** data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_STRING))
        return FALSE;
    *data=get_s_info(setting)->for_string.raw_values;
    return TRUE;
}

gboolean beryl_setting_get_int_max(BerylSetting * setting, gint * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_INT))
        return FALSE;
    *data=get_s_info(setting)->for_int.max;
    return TRUE;
}

gboolean beryl_setting_get_int_min(BerylSetting * setting, gint * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_INT))
        return FALSE;
    *data=get_s_info(setting)->for_int.min;
    return TRUE;
}

gboolean beryl_setting_get_float_min(BerylSetting * setting, gdouble * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_FLOAT))
        return FALSE;
    *data=get_s_info(setting)->for_float.min;
    return TRUE;
}

gboolean beryl_setting_get_float_max(BerylSetting * setting, gdouble * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_FLOAT))
        return FALSE;
    *data=get_s_info(setting)->for_float.max;
    return TRUE;
}

gboolean beryl_setting_get_float_precision(BerylSetting * setting, gdouble * data)
{
    if (!check_s_type(setting,BERYL_SETTING_TYPE_FLOAT))
        return FALSE;
    *data=get_s_info(setting)->for_float.precision;
    return TRUE;
}

gboolean beryl_setting_get_list_type(BerylSetting * setting, BerylSettingType * data)
{
    if (setting->type != BERYL_SETTING_TYPE_LIST)
        return FALSE;
    *data=setting->info.for_list.list_of_type;
    return TRUE;
}

BerylSettingType beryl_setting_get_type(BerylSetting * setting)
{
    return setting->type;
}

const gchar * beryl_setting_get_short_desc(BerylSetting * setting)
{
    return setting->short_desc;
}

const gchar * beryl_setting_get_long_desc(BerylSetting * setting)
{
    return setting->long_desc;
}

const gchar * beryl_setting_get_name(BerylSetting * setting)
{
    return setting->name;
}

const gchar * beryl_settings_plugin_get_short_desc(BerylSettingsPlugin * plugin)
{
    return plugin->short_desc;
}

const gchar * beryl_settings_plugin_get_long_desc(BerylSettingsPlugin * plugin)
{
    return plugin->long_desc;
}

const gchar * beryl_settings_plugin_get_name(BerylSettingsPlugin * plugin)
{
    return plugin->name;
}

GSList * beryl_settings_context_get_plugins(BerylSettingsContext * context)
{
    return context->plugins;
}

GSList * beryl_settings_plugin_get_settings(BerylSettingsPlugin * plugin)
{
    return plugin->settings;
}

gpointer beryl_settings_context_get_private(BerylSettingsContext * context)
{
    return context->private_ptr;
}

void beryl_settings_context_set_private(BerylSettingsContext * context, gpointer private_ptr)
{
    context->private_ptr = private_ptr;
}

gpointer beryl_settings_plugin_get_private(BerylSettingsPlugin * plugin)
{
    return plugin->private_ptr;
}

void beryl_settings_plugin_set_private(BerylSettingsPlugin * plugin, gpointer private_ptr)
{
    plugin->private_ptr = private_ptr;
}

gpointer beryl_setting_get_private(BerylSetting * setting)
{
    return setting->private_ptr;
}

void beryl_setting_set_private(BerylSetting * setting, gpointer private_ptr)
{
    setting->private_ptr = private_ptr;
}
