#define _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <libintl.h>
#include <beryl-settings-backend.h>

#include <gconf/gconf.h>
#include <gconf/gconf-client.h>
#include <gconf/gconf-value.h>

static GConfClient * client = NULL;

GConfValueType beryl_to_gconf_type(BerylSettingType bst);
// g_slist_nth_data(data,n) returns a gpointer, so * to dereference?
static void set_binding_value_from_gslist(BerylSettingValue * value, GSList *data, gint num_per, gint start)
{
    gboolean button_enabled;
    gboolean key_enabled;
    if (num_per==8)
    {
        //it's 8 if new-style semantics are used, allowing the enabled to be set separately
        //this should reintroduce some features gconf allowed but csm didn't
        button_enabled = g_slist_nth_data(data,start+6)?TRUE:FALSE;
        key_enabled = g_slist_nth_data(data,start+7)?TRUE:FALSE;
    }
    else if (num_per==6)
    {
        button_enabled = g_slist_nth_data(data, start + 1)?TRUE:FALSE;
        key_enabled = g_slist_nth_data(data, start + 3)?TRUE:FALSE;
    };

    if (num_per==6 || num_per==8)
    {
        beryl_setting_value_set_key_enabled(value,&key_enabled);
        beryl_setting_value_set_button_enabled(value,&button_enabled);
        beryl_setting_value_set_buttonmods(value,g_slist_nth_data(data, start + 0));
        beryl_setting_value_set_button(value,g_slist_nth_data(data, start + 1));
        beryl_setting_value_set_keymods(value,g_slist_nth_data(data, start + 2));
        beryl_setting_value_set_keysym(value,g_slist_nth_data(data, start + 3));
        beryl_setting_value_set_edgemask(value, g_slist_nth_data(data, start + 4));
        gboolean bell_set = g_slist_nth_data(data, start + 5)?TRUE:FALSE;
        beryl_setting_value_set_bell(value,&bell_set);
    };
}

static void set_color_value_from_gslist(BerylSettingValue *value, GSList *data, gint start)
{
    BerylSettingColorValue color_val;

    color_val.array.array[0] = (gint) g_slist_nth_data(data, start + 0);
    color_val.array.array[1] = (gint) g_slist_nth_data(data, start + 1);
    color_val.array.array[2] = (gint) g_slist_nth_data(data, start + 2);
    color_val.array.array[3] = (gint) g_slist_nth_data(data, start + 3);

    beryl_setting_value_set_color(value,&color_val);
}


static void set_gslist_from_binding(GSList ** plist, BerylSettingValue * value)
{
    GSList * list = NULL;
    gboolean bindval = value->value.as_binding.on_bell?1:0;
    gboolean bindbtn = value->value.as_binding.enabled.value.button?1:0;
    gboolean bindkey = value->value.as_binding.enabled.value.key?1:0;

    list = g_slist_append(list, &value->value.as_binding.button_mod_mask);
    list = g_slist_append(list, &value->value.as_binding.button);
    list = g_slist_append(list, &value->value.as_binding.key_mod_mask);
    list = g_slist_append(list, &value->value.as_binding.keysym);
    list = g_slist_append(list, &value->value.as_binding.edge_mask);
    list = g_slist_append(list, &bindval);
    list = g_slist_append(list, &bindbtn);
    list = g_slist_append(list, &bindkey);
    *plist = list;
}

static void set_gslist_from_color(GSList **plist, BerylSettingValue *value)
{
    GSList * list = NULL;
    list = g_slist_append(list, &value->value.as_color.array.array[0]);
    list = g_slist_append(list, &value->value.as_color.array.array[1]);
    list = g_slist_append(list, &value->value.as_color.array.array[2]);
    list = g_slist_append(list, &value->value.as_color.array.array[3]);
    *plist = list;
}


GConfValueType beryl_to_gconf_type(BerylSettingType bst)
{
    switch(bst)
    {
        case BERYL_SETTING_TYPE_BOOL:
            return GCONF_VALUE_BOOL;
        break;
        case BERYL_SETTING_TYPE_INT:
        case BERYL_SETTING_TYPE_COLOR:
        case BERYL_SETTING_TYPE_BINDING:
	    return GCONF_VALUE_INT;
        break;
        case BERYL_SETTING_TYPE_STRING:
            return GCONF_VALUE_STRING;
        break;
        case BERYL_SETTING_TYPE_FLOAT:
	    return GCONF_VALUE_FLOAT;
        break;
        default:
            return GCONF_VALUE_INVALID;
        break;
    }
}

void read_setting(BerylSettingsContext * context, BerylSetting * setting)
{
    const gchar * groupname;
    gchar * keyname;
    gchar *temp;
    gchar *gconf_path;
    gboolean bool_val;
    gint int_val;
    gdouble float_val;
    gchar * str_val;
    gsize ret_count;
    GSList *list = NULL;
    GError * e;

    beryl_setting_reset_to_default(setting);
    if (setting->parent->name)
        groupname=setting->parent->name;
    else
        groupname="_"; // pseudo-plugin for core

    // I /think/
    temp = g_strconcat("/apps/beryl/",groupname,NULL);
        groupname=g_strconcat(temp, "/", NULL);
    g_free(temp);

    if (setting->is_screen)
        keyname=g_strconcat("s_",setting->name,NULL);
    else
        keyname=g_strconcat("a_",setting->name,NULL);

    gconf_path = g_strconcat(groupname,keyname,NULL);

    if (!gconf_client_dir_exists(client,gconf_path, &e))
    {
        g_free(keyname);
	g_free(groupname);
        g_free(gconf_path);
        setting->is_default = TRUE;
        return;
    }

    e=NULL;

    setting->is_default=FALSE;
    switch(setting->type)
    {
        case BERYL_SETTING_TYPE_INT:
            int_val=gconf_client_get_int(client,gconf_path,&e);
            if (!e)
                beryl_setting_value_set_int(&setting->value,&int_val);
            break;
        case BERYL_SETTING_TYPE_FLOAT:
            float_val=gconf_client_get_float(client,gconf_path,&e);
            if (!e)
                beryl_setting_value_set_float(&setting->value,&float_val);
            break;
        case BERYL_SETTING_TYPE_BOOL:
            bool_val=gconf_client_get_bool(client,gconf_path,&e);
            if (!e)
                beryl_setting_value_set_bool(&setting->value,&bool_val);
            break;
        case BERYL_SETTING_TYPE_COLOR:
            list=gconf_client_get_list(client,gconf_path,GCONF_VALUE_INT,&e);
            if (!e && list && g_slist_length(list) == 4)
                    set_color_value_from_gslist(&setting->value,list, 0);

            g_slist_free(list);
            break;
        case BERYL_SETTING_TYPE_STRING:
            str_val=gconf_client_get_string(client,gconf_path,&e);
            if (!e && str_val)
            {
                //FIXME probably need _set_raw_string here
                //but since
                //beryl_setting_value_set_string(&setting->value,(const gchar **) &str_val);
                setting->value.value.as_string=g_strdup(str_val);
                setting->is_default=FALSE;
                g_free(str_val);
            }
            break;
        case BERYL_SETTING_TYPE_BINDING:
            list=gconf_client_get_list(client,gconf_path,GCONF_VALUE_INT,&e);
            if (!e && list)
                set_binding_value_from_gslist(&setting->value,list, g_slist_length(list), 0);

            g_slist_free(list);
            break;
        case BERYL_SETTING_TYPE_LIST:
            switch(setting->info.for_list.list_of_type)
            {
                int i;
                list = gconf_client_get_list(client,g_strconcat(groupname,keyname,NULL),beryl_to_gconf_type(setting->info.for_list.list_of_type),&e);
                ret_count = g_slist_length(list);

                if (e || !list)
                {
                    g_slist_free(list);
                    g_error_free(e);
		    break;
                };
                case BERYL_SETTING_TYPE_STRING:
                    beryl_setting_list_clear(setting);
                    for (i = 0;i < ret_count; i++)
                    {
                        BerylSettingValue * value=beryl_setting_list_append(setting);
                        value->value.as_string = g_strdup(g_slist_nth_data(list, i));
                        value->parent->is_default=FALSE;
                        //beryl_setting_value_set_string(value,(const gchar **) &str_ret[i]);
                    }

                    g_slist_free(list);
                    break;
                case BERYL_SETTING_TYPE_FLOAT:
                    beryl_setting_list_clear(setting);
                    for (i=0;i<ret_count;i++)
                    {
                        BerylSettingValue * value=beryl_setting_list_append(setting);
                        beryl_setting_value_set_float(value, g_slist_nth_data(list, i));
                    }

                    g_slist_free(list);
                    break;
                case BERYL_SETTING_TYPE_INT:
                    beryl_setting_list_clear(setting);
                    for (i=0;i<ret_count;i++)
                    {
                        BerylSettingValue * value=beryl_setting_list_append(setting);
                        beryl_setting_value_set_int(value,g_slist_nth_data(list, i));
                    }

                    g_slist_free(list);
                    break;
                case BERYL_SETTING_TYPE_BOOL:
                    beryl_setting_list_clear(setting);
                    for (i=0;i<ret_count;i++)
                    {
                        BerylSettingValue * value=beryl_setting_list_append(setting);
                        beryl_setting_value_set_bool(value,g_slist_nth_data(list, i));
                    }

                    g_slist_free(list);
                    break;
                case BERYL_SETTING_TYPE_COLOR:
                    if (ret_count%4==0)
                    {
                        beryl_setting_list_clear(setting);
                        for (i = 0; i < ret_count / 4; i++)
                        {
                            BerylSettingValue * value=beryl_setting_list_append(setting);
                            set_color_value_from_gslist(value,list, i*4);
                        }
                    }

                    g_slist_free(list);
                    break;
                case BERYL_SETTING_TYPE_BINDING:
                    if (ret_count%6==0 || ret_count%8==0)
                    {
                        beryl_setting_list_clear(setting);
                        int num_per=(ret_count%6==0)?6:8;
                        for (i=0;i<ret_count/num_per;i++)
                        {
                            BerylSettingValue * value=beryl_setting_list_append(setting);
                            set_binding_value_from_gslist(value, list, num_per, i*num_per);
                        }
                    }

                    g_slist_free(list);
                    break;
                default:
                    break;
            }
            break;
        default:
            break;
    }

    g_free(keyname);
    g_free(groupname);
    g_free(gconf_path);
}

static void append_string(BerylSettingValue * value, GSList ** sl)
{
    gchar * s;
    if (beryl_setting_value_get_string(value,&s))
    {
	*sl=g_slist_append(*sl,s);
    }
}

void write_setting(BerylSettingsContext * context, BerylSetting * setting)
{
    gchar *groupname=setting->parent->name?setting->parent->name:"_";
    groupname=g_strconcat(g_strconcat("/apps/beryl/",groupname,NULL), "/", NULL);
    gchar *keyname=g_strconcat(setting->is_screen?"s_":"a_",setting->name,NULL);
    gchar *gconf_path = g_strconcat(groupname,keyname, NULL);

    GError * e = NULL;

    if (setting->is_default)
    {
        gconf_client_unset(client,gconf_path,&e);
        g_free(keyname);
        g_free(groupname);
        g_free(gconf_path);
        return;
    }
    
    switch(setting->type)
    {
        case BERYL_SETTING_TYPE_STRING:
            gconf_client_set_string(client, gconf_path, setting->value.value.as_string,&e);
            break;
        case BERYL_SETTING_TYPE_FLOAT:
                gconf_client_set_float(client, gconf_path,setting->value.value.as_float,&e);
            break;
        case BERYL_SETTING_TYPE_INT:
            gconf_client_set_int(client,gconf_path, setting->value.value.as_int,&e);
            break;
        case BERYL_SETTING_TYPE_BOOL:
            gconf_client_set_bool(client,gconf_path, setting->value.value.as_bool,&e);
            break;
        case BERYL_SETTING_TYPE_BINDING:
            {
                GSList *list; 
                set_gslist_from_binding(&list,&setting->value);
                gconf_client_set_list(client,gconf_path, GCONF_VALUE_INT, list, &e);
                g_slist_free(list);
            }
            break;
        case BERYL_SETTING_TYPE_COLOR:
            {
		GSList *list; 
                set_gslist_from_color(&list, &setting->value);
                gconf_client_set_list(client, gconf_path,GCONF_VALUE_INT,list,&e);
                g_slist_free(list);
            }
            break;
        case BERYL_SETTING_TYPE_LIST:
	    switch(setting->info.for_list.list_of_type)
	    {
		case BERYL_SETTING_TYPE_STRING:
		    {
			GSList * sl = NULL;
			g_slist_foreach(setting->value.value.as_list,append_string,&sl);
			gconf_client_set_list(client,gconf_path,GCONF_VALUE_STRING,sl,&e);
			g_slist_free(sl);
		    }
		    break;
	    }
		//gconf_client_set_list(client, gconf_path, beryl_to_gconf_type(setting->info.for_list.list_of_type), setting->value.value.as_list, &e);
            break;
        default:
            break;
    }
    g_free(keyname);
    g_free(groupname);
    g_free(gconf_path);
    if (e)
	g_error_free(e);
}

gboolean read_init(BerylSettingsContext * context)
{
    GError *e = NULL;

    g_type_init();

    if (!client)
	client = gconf_client_get_default();

    if (!GCONF_IS_CLIENT(client))
    {
        printf("Error getting client :(\n");
        return FALSE;
    };
    if (gconf_client_dir_exists(client, "/apps/beryl", &e))
        gconf_client_add_dir(client,
		"/apps/beryl", GCONF_CLIENT_PRELOAD_RECURSIVE, &e);

    // Error-handling here!
    if (e) 
    {
        printf("gconf error preloading /apps/beryl\n");
        g_error_free(e);
        return FALSE;
    }
    return TRUE;
}

void read_done(BerylSettingsContext * context) {}

gboolean write_init(BerylSettingsContext * context)
{
    g_type_init();

    if (!client)
	client = gconf_client_get_default();

    gconf_client_add_dir(context->private_ptr,
	    "/apps/beryl/", GCONF_CLIENT_PRELOAD_NONE, NULL);
    return TRUE;
}

void write_done(BerylSettingsContext * context) {}
